/**
 * Yonius 0.3.21.
 *
 * Copyright (c) 2008-2020 Hive Solutions Lda.
 *
 * This source code is licensed under the Apache 2.0 license found in the
 * LICENSE file in the root directory of this source tree.
 */
import { promises } from 'fs';
import { join, normalize, resolve, dirname } from 'path';
import { env } from 'process';
import fetch from 'node-fetch';

var name$1 = "yonius";
var version = "0.3.21";

class Observable {
    constructor() {
        this.callbacks = {};
    }

    bind(event, callback) {
        const callbacks = this.callbacks[event] || [];
        callbacks.push(callback);
        this.callbacks[event] = callbacks;
        return callback;
    }

    unbind(event, callback) {
        const callbacks = this.callbacks[event] || [];
        if (!callback) {
            delete this.callbacks[event];
            return;
        }

        const index = callbacks.indexOf(callback);
        if (index === -1) {
            return;
        }
        callbacks.splice(index, 1);
        this.callbacks[event] = callbacks;
    }

    trigger(event) {
        const callbacks = this.callbacks[event] || [];
        const results = [];
        for (const callback of callbacks) {
            const result = callback.apply(this, Array.prototype.slice.call(arguments, 1));
            result !== undefined && result !== null && results.push(result);
        }
        return Promise.all(results);
    }
}

const verify = function(
    condition,
    message = null,
    code = null,
    exception = null,
    kwargs = {}
) {
    if (condition) return;
    message = message || "Verification failed";
    const Exception = exception || Error;
    kwargs = Object.assign({}, kwargs);
    if (message !== null && message !== undefined) kwargs.message = message;
    if (code !== null && message !== undefined) kwargs.code = code;
    const throwable = new Exception(kwargs.message || undefined);
    throwable.kwargs = kwargs;
    for (const [key, value] of Object.entries(kwargs)) {
        if (throwable[key] !== undefined) continue;
        throwable[key] = value;
    }
    throw throwable;
};

const verifyEqual = function(
    first,
    second,
    message = null,
    code = null,
    exception = null,
    kwargs = {}
) {
    message = message || `Expected ${first} got ${second}`;
    return this.verify(first === second, message, code, exception, kwargs);
};

const verifyNotEqual = function(
    first,
    second,
    message = null,
    code = null,
    exception = null,
    kwargs = {}
) {
    message = message || `Expected ${first} not equal to ${second}`;
    return this.verify(first !== second, message, code, exception, kwargs);
};

const verifyMany = function(
    sequence,
    message = null,
    code = null,
    exception = null,
    kwargs = {}
) {
    sequence.forEach(element => {
        verify(element, message, code, exception, kwargs);
    });
};

let HOME_DIR = null;

const pathExists = async function(path) {
    try {
        await promises.access(path);
    } catch (error) {
        return false;
    }
    return true;
};

const expandUser = function(path) {
    if (!path) return path;
    if (path === "~") return _homeDir();
    if (path.slice(0, 2) !== "~/") return path;
    return join(HOME_DIR, path.slice(2));
};

const _homeDir = function() {
    if (HOME_DIR !== null) return HOME_DIR;
    const isWindows = Boolean(process && process.platform === "win32");
    HOME_DIR = env[isWindows ? "USERPROFILE" : "HOME"] || "/";
    return HOME_DIR;
};

const FILE_NAME = "yonius.json";

const HOME_FILE = "~/.home";

const IMPORT_NAMES = ["$import", "$include", "$IMPORT", "$INCLUDE"];

const CASTS = {
    int: v => (typeof v === "number" ? v : parseInt(v)),
    float: v => (typeof v === "number" ? v : parseFloat(v)),
    bool: v => (typeof v === "boolean" ? v : ["1", "true", "True"].includes(v)),
    list: v => (Array.isArray(v) ? v : v.split(";")),
    tuple: v => (Array.isArray(v) ? v : v.split(";"))
};

const globals =
    typeof global === "undefined"
        ? typeof window === "undefined"
            ? typeof self === "undefined"
                ? {}
                : self
            : window
        : global;

globals.CONFIGS = globals.CONFIGS === undefined ? {} : globals.CONFIGS;

globals.CONFIG_F = globals.CONFIG_F === undefined ? [] : globals.CONFIG_F;

globals.HOMES = globals.HOMES === undefined ? [] : globals.HOMES;

globals.LOADED = globals.LOADED === undefined ? false : globals.LOADED;

const conf = function(name, fallback = undefined, cast = null, ctx = null) {
    const configs = ctx ? ctx.configs : globals.CONFIGS;
    cast = _castR(cast);
    let value = configs[name] === undefined ? fallback : configs[name];
    if (cast && value !== undefined && value !== null) value = cast(value);
    return value;
};

const confP = async function(name, fallback = undefined, cast = null, ctx = null) {
    await load();
    return conf(name, fallback, cast, ctx);
};

const confS = function(name, value, ctx = null) {
    const configs = ctx ? ctx.configs : globals.CONFIGS;
    configs[name] = value;
};

const load = async function(
    names = [FILE_NAME],
    path = null,
    encoding = "utf-8",
    force = false,
    ctx = null
) {
    if (globals.LOADED && !force) return;
    let paths = [];
    const homes = await getHomes();
    for (const home of homes) {
        paths = paths.concat([join(home), join(home, ".config")]);
    }
    if (path) paths.push(path);
    for (const path of paths) {
        for (const name of names) {
            await loadFile(name, path, encoding, ctx);
        }
    }
    await loadEnv(ctx);
    globals.LOADED = true;
};

const loadFile = async function(
    name = FILE_NAME,
    path = null,
    encoding = "utf-8",
    ctx = null
) {
    const configs = ctx ? ctx.configs : globals.CONFIGS;
    const configF = ctx ? ctx.configF : globals.CONFIG_F;

    let key;
    let value;
    let exists;
    let filePath;

    if (path) path = normalize(path);
    if (path) filePath = join(path, name);
    else filePath = name;

    filePath = resolve(filePath);
    filePath = normalize(filePath);
    const basePath = dirname(filePath);

    exists = await pathExists(filePath);
    if (!exists) return;

    exists = configF.includes(filePath);
    if (exists) configF.splice(configF.indexOf(filePath), 1);
    configF.push(filePath);

    const data = await promises.readFile(filePath, { encoding: encoding });
    const dataJ = JSON.parse(data);

    await _loadIncludes(basePath, dataJ, encoding);

    for ([key, value] of Object.entries(dataJ)) {
        if (!_isValid(key)) continue;
        configs[key] = value;
    }
};

const loadEnv = async function(ctx = null) {
    const configs = ctx ? ctx.configs : globals.CONFIGS;
    if (env === undefined || env === null) return;
    Object.entries(env).forEach(function([key, value]) {
        configs[key] = value;
    });
};

const getHomes = async function(
    filePath = HOME_FILE,
    fallback = "~",
    encoding = "utf-8",
    forceDefault = false
) {
    if (globals.HOMES.length > 0) return globals.HOMES;

    globals.HOMES = env.HOMES === undefined ? null : env.HOMES;
    globals.HOMES = globals.HOMES ? globals.HOMES.split(";") : globals.HOMES;
    if (globals.HOMES !== null) return globals.HOMES;

    fallback = expandUser(fallback);
    fallback = normalize(fallback);
    globals.HOMES = [fallback];

    filePath = expandUser(filePath);
    filePath = normalize(filePath);
    const exists = await pathExists(filePath);
    if (!exists) return globals.HOMES;

    if (!forceDefault) globals.HOMES.splice(0, globals.HOMES.length);

    let data = await promises.readFile(filePath, { encoding: encoding });
    data = data.trim();

    let paths = data.split(/\r?\n/);
    paths = paths.map(v => v.trim());

    for (let path of paths) {
        path = path.trim();
        if (!path) continue;
        path = expandUser(path);
        path = normalize(path);
        globals.HOMES.push(path);
    }

    return globals.HOMES;
};

const _castR = function(cast) {
    return CASTS[cast] === undefined ? cast : CASTS[cast];
};

const _loadIncludes = async function(basePath, config, encoding = "utf-8") {
    let includes = [];

    for (const alias of IMPORT_NAMES) {
        includes = config[alias] === undefined ? includes : config[alias];
    }

    if (typeof includes === "string") {
        includes = includes.split(";");
    }

    for (const include of includes) {
        await loadFile(include, basePath, encoding);
    }
};

const _isValid = function(key) {
    if (IMPORT_NAMES.includes(key)) return false;
    return true;
};

const _isDevel = function() {
    return ["DEBUG"].includes(conf("LEVEL", "INFO"));
};

const _isSecure = function() {
    return conf("SECURE", true, "bool");
};

/**
 * The map containing the various attribute alias between the normalized
 * manned and the appier manner.
 */
const ALIAS = {
    context: "find_d",
    filters: "find_d",
    "filters[]": "find_d",
    filter_def: "find_d",
    filter_string: "find_s",
    filter_name: "find_n",
    filter_operator: "find_o",
    insensitive: "find_i",
    order: "sort",
    offset: "skip",
    start_record: "skip",
    number_records: "limit"
};

/**
 * The map associating the various find fields with their respective
 * types, note that in case a special conversion operation is required
 * the associated value may represent a conversion function instead.
 */
const FIND_TYPES = {
    skip: v => parseInt(v),
    limit: v => Math.max(0, parseInt(v)),
    find_s: v => v,
    find_d: v => (Array.isArray(v) ? v : [v]),
    find_i: v => Boolean(v),
    find_t: v => v,
    find_n: v => v,
    find_o: v => v,
    sort: v => _toSort(v),
    meta: v => Boolean(v),
    fields: v => v
};

/**
 * The map that defines the various default values for a series of
 * find related attributes.
 */
const FIND_DEFAULTS = { limit: 10 };

/**
 * The map associating the normalized (text) way of representing sorting
 * with the current infra-structure number way of representing the same
 * information.
 */
const SORT_MAP = {
    1: 1,
    "-1": -1,
    ascending: 1,
    descending: -1
};

const getDataObject = function(
    params,
    alias = true,
    page = true,
    find = true,
    norm = true
) {
    let result = params;

    // in case the alias flag is set tries to resolve the attribute alias and
    // in case the find types are set converts the find based attributes using
    // the currently defined mapping map
    if (alias) result = _resolveAlias(result);
    if (page) result = _pageTypes(result);
    if (find) {
        result = _findTypes(result);
        result = _findDefaults(result);
    }

    // in case the normalization flag is set runs the normalization of the
    // provided object so that sequences are properly handled as defined in
    // the specification (this allows multiple references)
    if (norm) result = _normParams(result);

    // returns the constructed object to the caller method this object
    // should be a structured representation of the data in the request
    return result;
};

const _resolveAlias = function(params) {
    const result = {};
    Object.entries(params).forEach(([key, value]) => {
        result[ALIAS[key] || key] = value;
    });
    return result;
};

const _pageTypes = function(params, defaultSize = 50) {
    const result = Object.assign({}, params);

    const page = parseInt(params.page || 1);
    const size = parseInt(params.size || defaultSize);
    const offset = page - 1;
    result.skip = offset * size;
    result.limit = size;

    const sorter = params.sorter;
    const direction = params.direction || "descending";
    if (sorter) result.sort = `${sorter}:${direction}`;

    return result;
};

const _toSort = function(value) {
    const values = value.split(":", 2);
    if (values.length === 1) values.push("descending");
    const [name, direction] = values;
    if (name === "default") return null;
    return [name, SORT_MAP[direction] || 1];
};

const _findTypes = function(params) {
    const result = {};
    Object.entries(params).forEach(([key, value]) => {
        const converter = FIND_TYPES[key];
        const converted = converter ? converter(value) : value;
        result[key] = converted;
    });
    return result;
};

const _findDefaults = function(params) {
    const result = Object.assign({}, params);
    Object.entries(FIND_DEFAULTS).forEach(([key, value]) => {
        result[key] = params[key] || value;
    });
    return result;
};

const _normParams = function(params) {
    const result = Object.assign({}, params);

    // iterates over all the key value association in the object,
    // trying to find the ones that refer sequences so that they
    // may be normalized
    for (const [key, value] of Object.entries(params)) {
        // verifies if the current name references a sequence and
        // if that's not the case continues the loop trying to find
        // any other sequence based value
        if (!key.endsWith("[]")) {
            result[key] = value;
            continue;
        }

        // removes the extra sequence indication value
        const name = key.substring(0, key.length - 2);

        // in case the current value is not valid (empty) the object
        // is set with an empty list for the current iteration as this
        // is considered to be the default value
        if (!value) {
            result[name] = [];
            continue;
        }

        // retrieves the normalized and linearized list of leafs
        // for the current value and ten verifies the size of each
        // of its values and uses it to measure the number of
        // dictionary elements that are going to be contained in
        // the sequence to be "generated", then uses this (size)
        // value to pre-generate the complete set of dictionaries
        const leafs = _leafs(value);
        const [, values] = leafs[0] || [null, []];
        const list = values.map(_ => {});

        // sets the list of generates dictionaries in the object for
        // the newly normalized name of structure
        result[name] = list;

        // iterates over the complete set of key value pairs in the
        // leafs list to gather the value into the various objects that
        // are contained in the sequence (normalization process)
        for (const [name, value] of leafs) {
            for (let index; index < list.length; index++) {
                const object = list[index];
                const nameList = name.split(".");
                _setObject(object, nameList, value[index]);
            }
        }
    }

    return result;
};

/**
 * Retrieves a list containing a series of tuples that each represent a
 * leaf of the current object structure. A leaf is the last element of an
 * object that is not a map, the other intermediary maps are considered to
 * be trunks and should be percolated recursively.
 * This is a recursive function that takes some memory for the construction
 * of the list, and so should be used with the proper care to avoid bottlenecks.
 *
 * @param {Object} params The object for which the leafs list structure is
 * meant to be retrieved.
 *
 * @returns {Array} The list of leaf node tuples for the provided object,
 * as requested for each of the sequences.
 */
const _leafs = function(params) {
    // the list that will hold the various leaf nodes "gathered" by
    // the current recursion function
    let result = [];

    // iterates over all the key and value relations in the object trying
    // to find the leaf nodes (no map nodes) creating a tuple of fqn
    // (fully qualified name) and value
    for (const [key, value] of Object.entries(params)) {
        // retrieves the data type for the current value and validation
        // if it is a object or any other type in case it's an object a
        // new iteration step must be performed retrieving the leafs of
        // the value and then incrementing the name with the current prefix
        if (typeof value === "object") {
            const leafs = _leafs(value).map(([name, value]) => [`${key}.${name}`, value]);
            result = Array.concat(result, leafs);
        } else {
            // otherwise this is a leaf node and so the leaf tuple node
            // must be constructed with the current value (properly validated
            // for sequence presence)
            result.append([name, Array.isArray(value) ? value : [value]]);
        }
    }

    return result;
};

/**
 * Sets a composite value in an object, allowing for dynamic setting of
 * random size key values.
 * This method is useful for situations where one wants to set a value
 * at a randomly defined depth inside an object without having to much
 * work with the creation of the inner dictionaries.
 *
 * @param {Object} object The target object that is going to be
 * changed and set with the target value.
 * @param {Array} nameList The list of names that defined the fully
 * qualified name to be used in the setting of the value
 * for example path.to.end will be a three size list containing each
 * of the partial names.
 * @param {Object} value The value that is going to be set in the
 * defined target of the object.
 */
const _setObject = function(object, nameList, value) {
    // retrieves the first name in the names list this is the
    // value that is going to be used for the current iteration
    const [name, ...tail] = nameList[0];

    // in case the length of the current names list has reached
    // one this is the final iteration and so the value is set
    // at the current naming point
    if (nameList.length === 1) {
        object[name] = value;
    } else {
        // otherwise this is a "normal" step and so a new map must
        // be created/retrieved and the iteration step should be
        // performed on this new map as it's set on the current naming
        // place (recursion step)

        const map = object[name] || {};
        object[name] = map;
        _setObject(map, tail, value);
    }
};

class MixinBuilder {
    constructor(superclass) {
        this.superclass = superclass;
    }

    with(...mixins) {
        return mixins.reduce((c, mixin) => mixin(c), this.superclass);
    }
}

const mix = superclass => new MixinBuilder(superclass);

const matchOperatorsRegex = /[|\\{}()[\]^$+*?.-]/g;

const escapeStringRegexp = function(string) {
    if (typeof string !== "string") {
        throw new TypeError("Expected a string");
    }

    return string.replace(matchOperatorsRegex, "\\$&");
};

/**
 * The map that associates the various operators with the boolean
 * values that define if an insensitive base search should be used
 * instead of the "typical" sensitive search.
 */
const INSENSITIVE = {
    likei: true,
    llikei: true,
    rlikei: true
};

/**
 * The map containing the mapping association between the
 * normalized version of the operators and the infra-structure
 * specific value for each of this operations, note that some
 * of the values don't have a valid mapping for this operations
 * the operator must be ignored and not used explicitly.
 */
const OPERATORS = {
    eq: null,
    equals: null,
    ne: "$ne",
    not_equals: "$ne",
    in: "$in",
    nin: "$nin",
    not_in: "$nin",
    like: "$regex",
    likei: "$regex",
    llike: "$regex",
    llikei: "$regex",
    rlike: "$regex",
    rlikei: "$regex",
    gt: "$gt",
    greater: "$gt",
    gte: "$gte",
    greater_equal: "$gte",
    lt: "$lt",
    lesser: "$lt",
    lte: "$lte",
    lesser_equal: "$lte",
    null: null,
    is_null: null,
    not_null: "$ne",
    is_not_null: "$ne",
    contains: "$all"
};

/**
 * Map that associates each of the normalized operations with
 * an inline function that together with the data type maps the
 * the base string based value into the target normalized value.
 */
const VALUE_METHODS = {
    in: (v, t) => v.split(";").map(t),
    not_in: (v, t) => v.split(";").map(t),
    like: (v, t) => "^.*" + escapeStringRegexp(v) + ".*$",
    likei: (v, t) => "^.*" + escapeStringRegexp(v) + ".*$",
    llike: (v, t) => "^.*" + escapeStringRegexp(v) + "$",
    llikei: (v, t) => "^.*" + escapeStringRegexp(v) + "$",
    rlike: (v, t) => "^" + escapeStringRegexp(v) + ".*$",
    rlikei: (v, t) => "^" + escapeStringRegexp(v) + ".*$",
    null: (v, t) => null,
    is_null: (v, t) => null,
    not_null: (v, t) => null,
    is_not_null: (v, t) => null,
    contains: (v, t) => v.split(";")
};

const find = function(params, collection, modelClass) {
    _findS(params, modelClass);
    _findD(params, modelClass);

    const { skip = 0, limit = 0, sort = null } = params;

    collection.find({
        ...params,
        skip: skip,
        limit: limit,
        sort: sort
    });
};

const getDefault = function(modelClass) {
    return (
        (
            Object.entries(modelClass.schema).findOne(([name, definition]) => definition.default) ||
            {}
        ).name || null
    );
};

const getDefinitionN = function(name, modelClass = null) {
    if (modelClass === null) return {};
    return modelClass.schema[name];
};

const _findD = function(params, modelClass) {
    // retrieves the find definition into a local variable, then
    // removes the find definition from the named arguments map
    // so that it's not going to be erroneously used by the
    // underlying find infra-structure
    const findD = params.find_d;
    delete params.find_d;

    // in case the find definition is currently not defined in the
    // named arguments map returns immediately as nothing is
    // meant to be done on this method
    if (!findD) return;

    // tries to retrieve the value of the operator that is going
    // to be used to "join" the multiple find parts (find values)
    const findO = params.find_o;

    // verifies that the data type for the find definition is a
    // valid sequence and in case its not converts it into one
    // so that it may be used in sequence valid logic
    const _findD = Array.isArray(findD) ? findD : [findD];

    // iterates over all the filters defined in the filter definition
    // so that they may be used to update the provided arguments with
    // the filter defined in each of their lines
    for (const filter of _findD) {
        // in case the filter is not valid (unset or invalid) it's going
        // to be ignored as no valid information is present
        if (!filter) continue;

        // splits the filter string into its three main components
        // the name, operator and value, that are going to be processed
        // as defined by the specification to create the filter
        const result = filter.split(":", 3);
        if (result.length === 2) result.push(null);

        // unpacks the result into it's thee components name, operator
        // and value to be used in the parsing of the filter
        const [name, operator, value] = result;

        // retrieves the definition for the filter attribute and uses
        // it to retrieve it's target data type that is going to be
        // used for the proper conversion, note that in case the base
        // type resolution method exists it's used (recursive resolution)
        const nameDefinition = getDefinitionN(name, modelClass);
        const nameT = nameDefinition._btype || nameDefinition.type || String;

        // determines if the current filter operation should be performed
        // using a case insensitive based approach to the search, by default
        // all of the operations are considered to be case sensitive
        const insensitive = INSENSITIVE[operator] || false;

        // retrieves the method that is going to be used for value mapping
        // or conversion based on the current operator and then converts
        // the operator into the domain specific operator
        const valueMethod = VALUE_METHODS[operator];
        const _operator = OPERATORS[operator] === undefined ? operator : OPERATORS[operator];

        // in case there's a custom value mapped retrieved uses it to convert
        // the string based value into the target specific value for the query
        // otherwise uses the data type for the search field for value conversion
        const _value = valueMethod ? valueMethod(value, nameT) : nameT(value);

        // constructs the custom find value using a key and value map value
        // in case the operator is defined otherwise (operator not defined)
        // the value is used directly, then merges this find value into the
        // current set of filters for the provided (keyword) arguments
        let findV;
        if (_operator) {
            const obj = {};
            obj[_operator] = value;
            findV = obj;
        } else {
            findV = _value;
        }

        if (insensitive) findV.$options = "-i";
        _filterMerge(name, findV, params, findO);
    }
};

const _findS = function(params, modelClass) {
    // tries to retrieve the find name value from the provided
    // named arguments defaulting to an unset value otherwise
    const findN = params.find_n;
    delete params.find_n;

    // retrieves the kind of insensitive strategy that is going
    // to be used for the resolution of regular expressions,
    // this should affect all the filters and so it should be
    // used with some amount of care
    const findI = params.find_i || false;
    delete params.find_i;

    // retrieves the kind of default operation to be performed
    // this may be either: right, left or both and the default
    // value is both so that the token is matched in case it
    // appears anywhere in the search string
    const findT = params.find_t || "both";
    delete params.find_t;

    // retrieves the find string into a local variable, then
    // removes the find string from the named arguments map
    // so that it's not going to be erroneously used by the
    // underlying find infra-structure
    const findS = params.find_s;
    delete params.find_s;

    // in case the find string is currently not defined in the
    // named arguments map returns immediately as nothing is
    // meant to be done on this method
    if (!findS) return;

    // retrieves the "name" of the attribute that is considered
    // to be the default (representation) for the model in case
    // there's none returns immediately, as it's not possible
    // to proceed with the filter creation
    const defaultName = findN || getDefault(modelClass);
    if (!defaultName) return;

    // constructs the proper right and left parts of the regex
    // that is going to be constructed for the matching of the
    // value, this is achieved by checking the find type
    const right = findT === "right" ? "^" : "";
    const left = findT === "left" ? "$" : "";

    // retrieves the definition for the default attribute and uses
    // it to retrieve it's target data type, defaulting to the
    // string type in case none is defined in the schema
    const defaultT = getDefinitionN(defaultName, modelClass).type || String;

    let findV;

    try {
        // in case the target date type for the default field is
        // string the both sides wildcard regex is used for the
        // search
        if (defaultT === String) {
            findV = {
                $regex: right + escapeStringRegexp(findS) + left,
                $options: findI ? "-i" : ""
            };
        } else {
            findV = null;
        }
    } catch (err) {
        // in case there's an error in the conversion for
        // the target type value sets the search value as
        // invalid (not going to be used in filter)
        findV = null;
    }

    if (findV) _filterMerge(defaultName, findV, params);
};

const _filterMerge = function(name, filter, params, operator = null) {
    // retrieves a possible previous filter defined for the
    // provided name in case it does exist must concatenate
    // that previous value in a join statement according to
    // the currently defined operator
    const filterP = params[name];

    if (filterP || operator) {
        // defaults the operator for the join of the names to the
        // value and then ensures that the value of the operator
        // is within a valid range of values
        const _operator = operator || "$and";
        verify(["$and", "$or"].includes(_operator));

        // retrieves the and references for the current arguments
        // and appends the two filter values (current and previous)
        // then deletes the current name reference in the arguments
        // and updates the name value to the and value
        const filterA = params[_operator] || [];
        const _filter = filterP
            ? filterA.concat([{ name: filter }, { name: filterP }])
            : filterA.concat([{ name: filter }]);
        delete params[name];
        params[operator] = _filter;
    }

    params[name] = filter;
};

const equal = function(first, second) {
    if (first === second) {
        return true;
    }

    if (typeof_(first) !== typeof_(second)) {
        return false;
    }

    if (isPrimitive(first) && isPrimitive(second)) {
        return first === second;
    }

    if (Object.keys(first).length !== Object.keys(second).length) {
        return false;
    }

    for (const key in first) {
        if (!(key in second)) return false;
        if (!equal(first[key], second[key])) return false;
    }

    return true;
};

const isPrimitive = function(object) {
    return object !== Object(object);
};

const typeof_ = function(object) {
    if (object === null) return "null";
    if (Array.isArray(object)) return "array";
    return typeof object;
};

/* The default minimum value meaning that this is the
maximum value that one integer value may have for the
size rounding operation to be performed */
const DEFAULT_MINIMUM = 1024;

/* The default number of places (digits) that are going
to be used for the string representation in the round
based conversion of size units to be performed */
const DEFAULT_PLACES = 3;

/* The size unit coefficient as an integer value, this is
going to be used in each of the size steps as divisor */
const SIZE_UNIT_COEFFICIENT = 1024;

/* The simplified size units list that contains the complete set of
units indexed by the depth they represent */
const SIZE_UNITS_LIST_S = ["B", "K", "M", "G", "T", "P", "E", "Z", "Y"];

/* The size units list that contains the complete set of
units indexed by the depth they represent */
const SIZE_UNITS_LIST = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"];

const sizeRoundUnit = function(
    sizeValue,
    minimum = DEFAULT_MINIMUM,
    places = DEFAULT_PLACES,
    reduce = true,
    space = false,
    justify = false,
    simplified = false,
    depth = 0
) {
    // in case the current size value is acceptable (less than
    // the minimum) this is the final iteration and the final
    // string representation is going to be created
    if (sizeValue < minimum) {
        // calculates the maximum size of the string that is going
        // to represent the base size value as the number of places
        // plus one (representing the decimal separator character)
        const sizeS = places + 1;

        // calculates the target number of decimal places taking
        // into account the size (in digits) of the current size
        // value, this may never be a negative number
        const logValue = sizeValue && Math.log10(sizeValue);
        const digits = Math.trunc(logValue) + 1;
        places = places - digits;
        places = places > 0 ? places : 0;

        // rounds the size value, then converts the rounded
        // size value into a string based representation
        let sizeValueS = sizeValue.toFixed(places);

        // forces the reduce flag when the depth is zero, meaning
        // that an integer value will never be decimal, this is
        // required to avoid strange results for depth zero
        reduce = reduce || depth === 0;

        // in case the dot value is not present in the size value
        // string adds it to the end otherwise an issue may occur
        // while removing extra padding characters for reduce
        if (reduce && !sizeValueS.includes(".")) sizeValueS += ".";

        // strips the value from zero appended to the right and
        // then strips the value also from a possible decimal
        // point value that may be included in it, this is only
        // performed in case the reduce flag is enabled
        if (reduce) sizeValueS = sizeValueS.replace(/0+$/, "");
        if (reduce) sizeValueS = sizeValueS.replace(/\.$/, "");

        // in case the justify flag is set runs the justification
        // process on the size value taking into account the maximum
        // size of the associated size string
        if (justify) sizeValueS = _rjust(sizeValueS, sizeS);
        // retrieves the size unit (string mode) for the current
        // depth according to the provided map
        let sizeUnit;
        if (simplified) sizeUnit = SIZE_UNITS_LIST_S[depth];
        else sizeUnit = SIZE_UNITS_LIST[depth];

        // retrieves the appropriate separator based
        // on the value of the space flag
        const separator = (space && " ") || "";

        // creates the size value string appending the rounded
        // size value string and the size unit and returns it
        // to the caller method as the size value string
        const sizeValueString = sizeValueS + separator + sizeUnit;
        return sizeValueString;
    }
    // otherwise the value is not acceptable and a new iteration
    // must be ran with one less depth of size value
    else {
        // re-calculates the new size value, increments the depth
        // and runs the size round unit again with the new values
        const newSizeValue = parseFloat(sizeValue) / SIZE_UNIT_COEFFICIENT;
        const newDepth = depth + 1;
        return sizeRoundUnit(
            newSizeValue,
            minimum,
            places,
            reduce,
            space,
            justify,
            simplified,
            newDepth
        );
    }
};

const _rjust = function(str, length) {
    const lengthToFill = length - str.length;
    return lengthToFill <= 0 ? str : " ".repeat(lengthToFill) + str;
};

const camelToUnderscore = function(value, separator = "_") {
    if (!value) return value;
    return value
        .replace(/\.?([A-Z])/g, (x, y) => separator + y.toLowerCase())
        .replace(RegExp("^" + separator), "");
};

const urlEncode = function(values) {
    // constructs the parts array that is going to
    // store the multiple and values
    const parts = [];

    // in case the provided value is not an array
    // then assumes it's an object and retrieve entries
    if (!Array.isArray(values)) {
        values = Object.entries(values);
    }

    // iterates over the complete set of pairs available
    // from the key value pairs to be able to encode them
    // properly, notice that the values themselves can be
    // sequences allowing multiple repetition of key
    values.forEach(([key, value]) => {
        if (!Array.isArray(value)) {
            value = [value];
        }
        const keyEncoded = encodeURIComponent(key);
        value.forEach(_value => {
            if (_value === undefined || _value === null) {
                return;
            }
            const valueEncoded = encodeURIComponent(_value);
            parts.push(`${keyEncoded}=${valueEncoded}`);
        });
    });

    // joins the complete set of parts with the and
    // separator and then returns the final string value
    return parts.join("&");
};

const absoluteUrl = function(path, name = "BASE_URL", fallback = undefined) {
    const baseUrl = conf(name, fallback);
    return `${baseUrl}/${path}`;
};

class API extends Observable {
    constructor(kwargs = {}) {
        super();
        this.kwargs = kwargs;
    }

    async build(method, url, options = {}) {}

    async get(url, options = {}) {
        const result = await this._methodBasic("GET", url, options);
        return result;
    }

    async post(url, options = {}) {
        const result = await this._methodPayload("POST", url, options);
        return result;
    }

    async put(url, options = {}) {
        const result = await this._methodPayload("PUT", url, options);
        return result;
    }

    async delete(url, options = {}) {
        const result = await this._methodBasic("DELETE", url, options);
        return result;
    }

    async patch(url, options = {}) {
        const result = await this._methodPayload("PATCH", url, options);
        return result;
    }

    async options(url, options = {}) {
        const result = await this._methodBasic("OPTIONS", url, options);
        return result;
    }

    async _methodBasic(method, url, options = {}) {
        const params = options.params !== undefined ? options.params : {};
        const headers = options.headers !== undefined ? options.headers : {};
        const kwargs = options.kwargs !== undefined ? options.kwargs : {};
        const handle = options.handle !== undefined ? options.handle : true;
        await this.build(method, url, {
            params: params,
            headers: headers,
            kwargs: kwargs
        });
        const query = urlEncode(params || {});
        if (query) url += url.includes("?") ? "&" + query : "?" + query;
        const response = await fetch(url, {
            method: method,
            headers: headers || {}
        });
        const result = handle ? await this._handleResponse(response) : response;
        return result;
    }

    async _methodPayload(method, url, options = {}) {
        const params = options.params !== undefined ? options.params : {};
        let headers = options.headers !== undefined ? options.headers : {};
        let data = options.data !== undefined ? options.data : null;
        const dataJ = options.dataJ !== undefined ? options.dataJ : null;
        const dataM = options.dataM !== undefined ? options.dataM : null;
        let mime = options.mime !== undefined ? options.mime : null;
        const kwargs = options.kwargs !== undefined ? options.kwargs : {};
        const handle = options.handle !== undefined ? options.handle : true;

        await this.build(method, url, {
            params: params,
            headers: headers,
            data: data,
            dataJ: dataJ,
            dataM: dataM,
            mime: mime,
            kwargs: kwargs
        });

        const query = urlEncode(params || {});

        if (data !== null) {
            if (query) url += url.includes("?") ? "&" + query : "?" + query;
        } else if (dataJ !== null) {
            data = JSON.stringify(dataJ);
            if (query) url += url.includes("?") ? "&" + query : "?" + query;
            mime = mime || "application/json";
        } else if (dataM !== null) {
            if (query) url += url.includes("?") ? "&" + query : "?" + query;
        } else if (query) {
            data = query;
            mime = mime || "application/x-www-form-urlencoded";
        }

        headers = Object.assign({}, headers);
        if (mime) headers["Content-Type"] = mime;

        const response = await fetch(url, {
            method: method,
            headers: headers || {},
            body: data
        });
        const result = handle ? await this._handleResponse(response) : response;
        return result;
    }

    async _handleResponse(response, errorMessage = "Problem in request") {
        let result = null;
        if (response.headers.get("content-type").toLowerCase().startsWith("application/json")) {
            result = await response.json();
        } else {
            result = await response.blob();
        }
        verify(response.ok, result.error || errorMessage, response.status || 500);
        return result;
    }
}

class OAuthAPI extends API {}

class OAuth1API extends OAuthAPI {}

class OAuth2API extends OAuthAPI {
    constructor(kwargs = {}) {
        super(kwargs);
        this.accessToken = null;
    }

    async build(method, url, options = {}) {
        await super.build(method, url, options);
        const params = options.params !== undefined ? options.params : {};
        const headers = options.headers !== undefined ? options.headers : {};
        const kwargs = options.kwargs !== undefined ? options.kwargs : {};
        const token = kwargs.token === undefined ? this.tokenDefault : kwargs.token;
        delete kwargs.token;
        if (token && this.oauthTypes.includes("param")) {
            params[this.oauthParam] = this.getAccessToken();
        }
        if (token && this.oauthTypes.includes("header")) {
            headers.Authorization = `Bearer ${this.getAccessToken()}`;
        }
    }

    getAccessToken() {
        if (this.accessToken) return this.accessToken;
        throw new Error("No access token found must re-authorize");
    }

    get oauthTypes() {
        return ["param", "header"];
    }

    get oauthParam() {
        return "access_token";
    }

    get tokenDefault() {
        return true;
    }
}

class YoniusError extends Error {
    constructor(message, code = 500) {
        super(message);
        this.code = code;
    }

    get isClient() {
        return Math.floor(this.code / 100) === 4;
    }

    get isServer() {
        return Math.floor(this.code / 100) === 5;
    }
}

class OperationalError extends YoniusError {
    constructor(message = "Operational error", code = 500) {
        super(message, code);
    }
}

class NotFoundError extends OperationalError {
    constructor(message = "Not found", code = 404) {
        super(message, code);
    }
}

class NotImplementedError extends OperationalError {
    constructor(message = "Not implemented", code = 501) {
        super(message, code);
    }
}

/**
 * Ensures that the current "session" context contains the
 * requested ACL token as valid.
 *
 * In case the validation fails an exception is raised
 * indicating the auth validation error.
 *
 * @param {String} token The ACL token to ensure permission,
 * the logged user should be allowed to id.
 * @param {Object} ctx The context object to be used in
 * the session basic ACL retrieval, should contain proper
 * injected methods for retrieval (eg: `getAcl`).
 */
const ensurePermissions = async (token, ctx) => {
    // retrieves the ACL values from the current context and
    // then uses the ACL to obtain the valid expanded tokens map
    const acl = ctx.getAcl ? await ctx.getAcl(ctx) : {};
    const tokens = toTokensM(acl);

    // in case the permission validation test is not positive
    // then an exception should be raised indicating the issue
    if (!_hasPermission(token, tokens)) {
        throw new OperationalError("You don't have authorization to access this resource", 401);
    }
};

/**
 * Converts the provided list of token strings separated by dots
 * into a map based representation on an hierarchical structure.
 *
 * @param {Array} tokens A linear array of tokens to convert into
 * an hierarchical representation.
 * @returns {Object} The map containing the hierarchy of tokens
 * for the provided linear string based sequence of tokens.
 */
const toTokensM = tokens => {
    const tokensM = {};

    if (tokens === undefined) return tokensM;
    if (tokens === null) return tokensM;
    if (!Array.isArray(tokens)) return tokensM;

    for (const token of tokens) {
        let tokensC = tokensM;
        const tokenL = token.split(".");
        const head = tokenL.slice(0, tokenL.length - 1);
        const tail = tokenL[tokenL.length - 1];

        for (const tokenP of head) {
            let current = tokensC[tokenP] || {};
            const isDict = typeof current === "object";
            if (!isDict) current = { _: current };
            tokensC[tokenP] = current;
            tokensC = current;
        }

        const leaf = tokensC[tail] || null;
        if (leaf && typeof leaf === "object") leaf._ = true;
        else tokensC[tail] = true;
    }

    return tokensM;
};

const _hasPermission = (token, tokensM = null) => {
    if (!token) return true;
    if (tokensM === undefined || tokensM === null) return false;

    const tokenL = token.split(".");
    for (const tokenP of tokenL) {
        if (typeof tokensM !== "object") return false;
        if (tokensM["*"]) return true;
        if (tokensM[tokenP] === undefined) return false;
        tokensM = tokensM[tokenP];
    }

    const isDict = typeof tokensM === "object";
    const result = isDict ? tokensM._ || false : tokensM;

    return Boolean(result);
};

const REGISTRY = {};

const load$1 = async function() {
    await load();
};

const unload = async function() {};

const register = function(name, value) {
    REGISTRY[name] = value;
};

const unregister = function(name) {
    delete REGISTRY[name];
};

const request = function(name) {
    if (REGISTRY[name] === undefined) {
        throw new OperationalError("Name not found in registry");
    }
    return REGISTRY[name];
};

class Collection {
    constructor(options) {
        this.options = options;
    }

    async find(conditions, projection = {}, options = {}) {
        throw new NotImplementedError();
    }

    async findOne(conditions, projection = {}, options = {}) {
        throw new NotImplementedError();
    }

    async findOneAndUpdate(conditions, data, options = {}) {
        throw new NotImplementedError();
    }

    async findOneAndDelete(conditions, options = {}) {
        throw new NotImplementedError();
    }

    async create(data, options = {}) {
        throw new NotImplementedError();
    }
}

class MongoCollection extends Collection {
    constructor(name, schema) {
        const mongoose = request("mongoose");
        super(name, schema);
        this._mongoose = mongoose.model(
            this.options.name,
            new mongoose.Schema(this.options.schema)
        );
    }

    async find(conditions, projection = {}, options = {}) {
        const model = await this._mongoose.find(conditions, projection, options);
        return model;
    }

    async findOne(conditions, projection = {}, options = {}) {
        const model = this._mongoose.findOne(conditions, projection, options);
        return model;
    }

    async findOneAndUpdate(conditions, data, options = {}) {
        const model = await this._mongoose.findOneAndUpdate(conditions, data, {
            upsert: true,
            new: true,
            ...options
        });
        return model;
    }

    async findOneAndDelete(conditions, options = {}) {
        const model = await this._mongoose.findOneAndDelete(conditions, options);
        return model;
    }

    async create(data, options = {}) {
        const models = await this._mongoose.create([data], options);
        return models[0];
    }
}

var collection = /*#__PURE__*/Object.freeze({
    __proto__: null,
    Collection: Collection,
    MongoCollection: MongoCollection,
    'default': Collection
});

const MEMORY_STORAGE = {};

class Model {
    static niw() {
        return new this();
    }

    static get adapter() {
        return process.env.ADAPTER || "mongo";
    }

    async validate() {
        const errors = [...this._validate()];
        if (errors.length) {
            throw new Error(`Invalid model: ${errors.map(err => String(err)).join(", ")}`);
        }
    }

    async apply(model) {
        await this.wrap(model);
        return this;
    }

    async wrap(model) {
        await this._wrap(model);
        return this;
    }

    get model() {
        return this;
    }

    get jsonV() {
        return this.model;
    }

    get string() {
        return JSON.stringify(this.model);
    }

    async _wrap(model) {
        for (const key of Object.keys(this.constructor.schema)) {
            if (model[key] === undefined) continue;
            this[key] = model[key];
        }
        if (model._id !== undefined) this._id = model._id;
    }

    * _validate() {}
}

class ModelStore extends Model {
    static async find(options = {}) {
        const found = await this.collection.find(options);
        const models = await Promise.all(found.map(v => new this().wrap(v)));
        return models;
    }

    static async get(id) {
        const options = {};
        options[this.idName] = id;
        const model = await this.collection.findOne(options);

        if (!model) {
            throw new NotFoundError();
        }

        const instance = new this();
        return instance.wrap(model);
    }

    static get schema() {
        throw new NotImplementedError();
    }

    static get collection() {
        if (this._collection) return this._collection;
        const adapter = this.adapter[0].toUpperCase() + this.adapter.slice(1);
        this._collection = new collection[adapter + "Collection"](this.dataOptions);
        return this._collection;
    }

    static get idName() {
        return "id";
    }

    static get dataOptions() {
        return {
            name: this.name,
            schema: this.schema
        };
    }

    async save(validate = true) {
        let model;
        if (validate) await this.validate();
        await this.verify();
        const isNew = this._id === undefined;
        const conditions = {};
        conditions[this.constructor.idName] = this.identifier;
        if (isNew) model = await this.constructor.collection.create(this.model);
        else model = await this.constructor.collection.findOneAndUpdate(conditions, this.model);
        this.wrap(model);
        return this;
    }

    async delete() {
        const conditions = {};
        conditions[this.constructor.idName] = this.identifier;
        await this.constructor.collection.findOneAndDelete(conditions);
        return this;
    }

    async verify() {
        verify(
            this.identifier !== undefined && this.identifier !== null,
            "The identifier must be defined before saving"
        );
        for (const [name, field] of Object.entries(this.constructor.schema)) {
            verify(
                !field.required || ![undefined, null].includes(this[name]),
                `No value provided for mandatory field '${name}'`
            );
        }
    }

    get identifier() {
        return this[this.constructor.idName];
    }
}

class ModelMemory extends ModelStore {
    static get adapter() {
        return "memory";
    }

    static get dataOptions() {
        return Object.assign(super.dataOptions, { storage: this.storage });
    }

    static get storage() {
        return MEMORY_STORAGE[this.name];
    }
}

const ensureExpress = token => {
    return (req, res, next) => {
        ensurePermissions(token, req).catch(next).then(next);
    };
};

const NAME = name$1;
const VERSION = version;

export { ALIAS, API, Collection, FIND_DEFAULTS, FIND_TYPES, INSENSITIVE, MixinBuilder, Model, ModelMemory, ModelStore, MongoCollection, NAME, NotFoundError, NotImplementedError, OAuth1API, OAuth2API, OAuthAPI, OPERATORS, Observable, OperationalError, SORT_MAP, VALUE_METHODS, VERSION, YoniusError, _castR, _isDevel, _isSecure, _isValid, absoluteUrl, camelToUnderscore, conf, confP, confS, ensureExpress, ensurePermissions, equal, escapeStringRegexp, expandUser, find, getDataObject, globals, isPrimitive, load$1 as load, load as loadConf, mix, pathExists, register, request, sizeRoundUnit, toTokensM, typeof_, unload, unregister, urlEncode, verify, verifyEqual, verifyMany, verifyNotEqual };
//# sourceMappingURL=yonius.esm.js.map
